/*
 * xvkbd - Virtual Keyboard for X Window System
 * (Version 1.0, 2000-11-03)
 *
 * Copyright (C) 2000 by Tom Sato <VEF00200@nifty.ne.jp>
 * http://member.nifty.ne.jp/tsato/
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 */

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>

#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Shell.h>
#include <X11/keysym.h>
#include <X11/cursorfont.h>
#include <X11/Xproto.h>  /* to get request code */
#include <X11/Xaw/Form.h>
#include <X11/Xaw/Command.h>
#include <X11/Xaw/Repeater.h>
#include <X11/Xaw/Label.h>
#include <X11/Xmu/WinUtil.h>

#include "resources.h"
#define PROGRAM_NAME_WITH_VERSION "xvkbd (v1.0)"

/*
 * In this release, keyboard layout is hardcoded here.
 */
#define NUM_KEY_ROWS     6
#define NUM_KEY_COLS    16

static char *keys_normal[NUM_KEY_ROWS][NUM_KEY_COLS] = {
  { "F1", "F2", "F3", "F4", "F5", "F6", "F7", "F8", "F9", "F10", "F11", "F12", "BackSpace" },
  { "Escape", "1", "2", "3", "4", "5", "6", "7", "8", "9", "0", "-", "=", "\\", "`" },
  { "Tab", "q", "w", "e", "r", "t", "y", "u", "i", "o", "p", "[", "]", "Delete" },
  { "Control_L", "a", "s", "d", "f", "g", "h", "j", "k", "l", ";", "'", "Return" },
  { "Shift_L", "z", "x", "c", "v", "b", "n", "m", ",", ".", "/", "Multi_key", "Shift_R" },
  { "Caps_Lock", "Alt_L", "Meta_L", "space", "Meta_R", "Alt_R", "Left", "Right", "Up", "Down" },
};
static char *keys_shift[NUM_KEY_ROWS][NUM_KEY_COLS] = {
  { "F1", "F2", "F3", "F4", "F5", "F6", "F7", "F8", "F9", "F10", "F11", "F12", "BackSpace" },
  { "Escape", "!", "@", "#", "$", "%", "^", "&", "*", "(", ")", "_", "+", "|", "~" },
  { "Tab", "Q", "W", "E", "R", "T", "Y", "U", "I", "O", "P", "{", "}", "Delete" },
  { "Control_L", "A", "S", "D", "F", "G", "H", "J", "K", "L", ":", "\"", "Return" },
  { "Shift_L", "Z", "X", "C", "V", "B", "N", "M", "<", ">", "?", "Multi_key", "Shift_R" },
  { "Caps_Lock", "Alt_L", "Meta_L", "space", "Meta_R", "Alt_R", "Left", "Right", "Up", "Down" },
};
static char *keys_altgr[NUM_KEY_ROWS][NUM_KEY_COLS] = {};
static char *keys_shift_altgr[NUM_KEY_ROWS][NUM_KEY_COLS] = {};

static char *key_labels[NUM_KEY_ROWS][NUM_KEY_COLS] = {
  { "F1", "F2", "F3", "F4", "F5", "F6", "F7", "F8", "F9", "F10", "F11", "F12", "Backspace" },
  { "Esc", "!\n1", "@\n2", "#\n3", "$\n4", "%\n5", "^\n6",
    "&\n7", "*\n8", "(\n9", ")\n0", "_\n-", "+\n=", "|\n\\", "~\n`" },
  { "Tab", "Q", "W", "E", "R", "T", "Y", "U", "I", "O", "P", "{\n[", "}\n]", "Del" },
  { "Control", "A", "S", "D", "F", "G", "H", "J", "K", "L", ":\n;", "\"\n'", "Return" },
  { "Shift", "Z", "X", "C", "V", "B", "N", "M", "<\n,", ">\n.", "?\n/", "Com\npose", "Shift" },
  { "Caps\nLock", "Alt", "Meta", "", "Meta", "Alt", "left", "right", "up", "down" },
};

#define NUM_KEYPAD_ROWS  5
#define NUM_KEYPAD_COLS  5

static char *keypad[NUM_KEYPAD_ROWS][NUM_KEYPAD_COLS] = {
  { "Print",     "Scroll_Lock", "Pause",       "Focus"       },
  { "Home",      "Up",          "Page_Up",     "KP_Add"      },
  { "Left",      "5",           "Right",       "KP_Subtract" },
  { "End",       "Down",        "Page_Down",   "KP_Enter"    },
  { "Insert",                   "Delete"                     },
};
static char *keypad_shift[NUM_KEYPAD_ROWS][NUM_KEYPAD_COLS] = {
  { "KP_Equal",  "KP_Divide",   "KP_Multiply", "Focus"       },
  { "KP_7",      "KP_8",        "KP_9",        "KP_Add"      },
  { "KP_4",      "KP_5",        "KP_6",        "KP_Subtract" },
  { "KP_1",      "KP_2",        "KP_3",        "KP_Enter"    },
  { "KP_0",                     "."                          },
};
static char *keypad_label[NUM_KEYPAD_ROWS][NUM_KEYPAD_COLS] = {
  { "=\nPrint",  "/\nScrLk",    "*\nPause",    "Focus"       },
  { "7\nHome",   "8\nUp  ",     "9\nPgUp",     "+"           },
  { "4\nLeft",   "5\n    ",     "6\nRight",    "-"           },
  { "1\nEnd ",   "2\nDown",     "3\nPgDn",     "Enter"       },
  { "0\nIns ",                  ".\nDel "                    },
};

/*
 * Image for arrow keys
 */
#define up_width 7
#define up_height 13
static unsigned char up_bits[] = {
   0x08, 0x1c, 0x1c, 0x3e, 0x2a, 0x49, 0x08, 0x08, 0x08, 0x08, 0x08, 0x08,
   0x08};

#define down_width 7
#define down_height 13
static unsigned char down_bits[] = {
   0x08, 0x08, 0x08, 0x08, 0x08, 0x08, 0x08, 0x49, 0x2a, 0x3e, 0x1c, 0x1c,
   0x08};

#define left_width 13
#define left_height 13
static unsigned char left_bits[] = {
   0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x20, 0x00, 0x18, 0x00, 0x0e, 0x00,
   0xff, 0x1f, 0x0e, 0x00, 0x18, 0x00, 0x20, 0x00, 0x00, 0x00, 0x00, 0x00,
   0x00, 0x00};

#define right_width 13
#define right_height 13
static unsigned char right_bits[] = {
   0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x80, 0x00, 0x00, 0x03, 0x00, 0x0e,
   0xff, 0x1f, 0x00, 0x0e, 0x00, 0x03, 0x80, 0x00, 0x00, 0x00, 0x00, 0x00,
   0x00, 0x00};

/*
 * Resources and options
 */
#define Offset(entry) XtOffset(struct appres_struct *, entry)
static XtResource application_resources[] = {
  { "windowGeometry", "Geometry", XtRString, sizeof(char *),
    Offset(geometry), XtRImmediate, "" },
  { "debug", "Debug", XtRBoolean, sizeof(Boolean),
     Offset(debug), XtRImmediate, (XtPointer)FALSE },
  { "kterm", "Kterm", XtRBoolean, sizeof(Boolean),
     Offset(kterm), XtRImmediate, (XtPointer)FALSE },
  { "autoRepeat", "AutoRepeat", XtRBoolean, sizeof(Boolean),
     Offset(auto_repeat), XtRImmediate, (XtPointer)TRUE },
  { "keypad", "Keypad", XtRBoolean, sizeof(Boolean),
     Offset(keypad), XtRImmediate, (XtPointer)TRUE },
  { "functionkey", "FunctionKey", XtRBoolean, sizeof(Boolean),
     Offset(function_key), XtRImmediate, (XtPointer)TRUE },
  { "compact", "Compact", XtRBoolean, sizeof(Boolean),
     Offset(compact), XtRImmediate, (XtPointer)FALSE },
  { "keypadOnly", "KeypadOnly", XtRBoolean, sizeof(Boolean),
     Offset(keypad_only), XtRImmediate, (XtPointer)FALSE },
  { "keypadKeysym", "KeypadKeysym", XtRBoolean, sizeof(Boolean),
     Offset(keypad_keysym), XtRImmediate, (XtPointer)FALSE },
  { "listWidgets", "Debug", XtRBoolean, sizeof(Boolean),
     Offset(list_widgets), XtRImmediate, (XtPointer)FALSE },
  { "text", "Text", XtRString, sizeof(char *),
    Offset(text), XtRImmediate, "" },
  { "window", "Window", XtRString, sizeof(char *),
    Offset(window), XtRImmediate, "" },
  { "widget", "Widget", XtRString, sizeof(char *),
    Offset(widget), XtRImmediate, "" },
  { "generalFont", XtCFont, XtRFontStruct, sizeof(XFontStruct *),
      Offset(general_font), XtRString, XtDefaultFont},
  { "letterFont", XtCFont, XtRFontStruct, sizeof(XFontStruct *),
      Offset(letter_font), XtRString, XtDefaultFont},
  { "specialFont", XtCFont, XtRFontStruct, sizeof(XFontStruct *),
      Offset(special_font), XtRString, XtDefaultFont},
  { "keypadFont", XtCFont, XtRFontStruct, sizeof(XFontStruct *),
      Offset(keypad_font), XtRString, XtDefaultFont},
  { "generalBackground", XtCBackground, XtRPixel, sizeof(Pixel),
     Offset(general_background), XtRString, "gray" },
  { "specialBackground", XtCBackground, XtRPixel, sizeof(Pixel),
     Offset(special_background), XtRString, "gray" },
  { "specialForeground", XtCForeground, XtRPixel, sizeof(Pixel),
     Offset(special_foreground), XtRString, "black" },
  { "highlightBackground", XtCBackground, XtRPixel, sizeof(Pixel),
     Offset(highlight_background), XtRString, "gray" },
  { "highlightForeground", XtCForeground, XtRPixel, sizeof(Pixel),
     Offset(highlight_foreground), XtRString, "forestgreen" },
  { "focusBackground", XtCBackground, XtRPixel, sizeof(Pixel),
     Offset(focus_background), XtRString, "gray" },
  { "normalkeys", "NormalKeys", XtRString, sizeof(char *),
    Offset(keys_normal), XtRImmediate, "" },
  { "shiftkeys", "ShiftKeys", XtRString, sizeof(char *),
    Offset(keys_shift), XtRImmediate, "" },
  { "altgrkeys", "AltgrKeys", XtRString, sizeof(char *),
    Offset(keys_altgr), XtRImmediate, "" },
  { "shiftaltgrkeys", "ShiftAltgrKeys", XtRString, sizeof(char *),
    Offset(keys_shift_altgr), XtRImmediate, "" },
  { "keylabels", "KeyLabels", XtRString, sizeof(char *),
    Offset(key_labels), XtRImmediate, "" },
};
#undef Offset

static XrmOptionDescRec options[] = {
  { "-geometry", ".windowGeometry", XrmoptionSepArg, NULL },
  { "-windowgeometry", ".windowGeometry", XrmoptionSepArg, NULL },
  { "-debug", ".debug", XrmoptionNoArg, "True" },
  { "-kterm", ".kterm", XrmoptionNoArg, "True" },
  { "-no-kterm", ".kterm", XrmoptionNoArg, "False" },
  { "-nokterm", ".kterm", XrmoptionNoArg, "False" },
  { "-text", ".text", XrmoptionSepArg, NULL },
  { "-window", ".window", XrmoptionSepArg, NULL },
  { "-widget", ".widget", XrmoptionSepArg, NULL },
  { "-no-repeat", ".autoRepeat", XrmoptionNoArg, "False" },
  { "-norepeat", ".autoRepeat", XrmoptionNoArg, "False" },
  { "-no-keypad", ".keypad", XrmoptionNoArg, "False" },
  { "-nokeypad", ".keypad", XrmoptionNoArg, "False" },
  { "-no-functionkey", ".functionkey", XrmoptionNoArg, "False" },
  { "-nofunctionkey", ".functionkey", XrmoptionNoArg, "False" },
  { "-highlight", ".highlightForeground", XrmoptionSepArg, NULL },
  { "-compact", ".compact", XrmoptionNoArg, "True" },
  { "-keypad", ".keypadOnly", XrmoptionNoArg, "True" },
  { "-true-keypad", ".keypadKeysym", XrmoptionNoArg, "True" },
  { "-truekeypad", ".keypadKeysym", XrmoptionNoArg, "True" },
  { "-list", ".listWidgets", XrmoptionNoArg, "True" },
};

/*
 * Global variables
 */
static XtAppContext app_con;
static Widget toplevel;
static Display *dpy;
static Atom wm_delete_window = None;

static KeySym *keysym_table;
static int min_keycode, max_keycode;
static int keysym_per_keycode;
static Boolean error_detected;

static XModifierKeymap *modifiers;
static int alt_mask = 0;
static int meta_mask = 0;
static int mode_sw_mask = 0;

static int shift_state = 0;
static int mouse_shift = 0;

static Window toplevel_parent = None;
static Window focused_window = None;
static Window focused_subwindow = None;

static Window *focused_children = NULL;
static unsigned int n_focused_children = 0;

/*
 * Search for window which has specified instance name (WM_NAME)
 * or class name (WM_CLASS).
 */
static Window FindWindow(Window top, char *name)
{
  Window w;
  Window *children, dummy;
  unsigned int nchildren;
  int i;
  XClassHint hint;

  w = None;

  if (appres.debug) fprintf(stderr, "FindWindow: id=0x%lX", (long)top);

  if (XGetClassHint(dpy, top, &hint)) {
    if (hint.res_name) {
      if (appres.debug) fprintf(stderr, " instance=\"%s\"", hint.res_name);
      if (strcmp(hint.res_name, name) == 0) w = top;
      XFree(hint.res_name);
    }
    if (hint.res_class) {
      if (appres.debug) fprintf(stderr, " class=\"%s\"", hint.res_class);
      if (strcmp(hint.res_class, name) == 0) w = top;
      XFree(hint.res_class);
    }
  }
  if (appres.debug) fprintf(stderr, "\n");

  if (w == None &&
      XQueryTree(dpy, top, &dummy, &dummy, &children, &nchildren)) {
    for (i = 0; i < nchildren; i++) {
      w = FindWindow(children[i], name);
      if (w != None) break;
    }
    if (children) XFree((char *)children);
  }

  return(w);
}

/*
 * This will be called to get window to set input focus,
 * when user pressed the "Focus" button.
 */
static void GetFocusedWindow(void)
{
  Cursor crosshair_cursor;
  XEvent event;
  Window root, parent, child;
  int junk_i;
  unsigned junk_u;

  crosshair_cursor = XCreateFontCursor(dpy, XC_crosshair);

  root = RootWindow(dpy, DefaultScreen(dpy));
  if (XGrabPointer(dpy, root, False, ButtonPressMask,
                   GrabModeSync, GrabModeAsync, None,
                   crosshair_cursor, CurrentTime) == 0) {
    if (appres.debug)
      fprintf(stderr, "Grab pointer - waiting for button press\n");
    do {
      XAllowEvents(dpy, SyncPointer, CurrentTime);
      XNextEvent(dpy, &event);
    } while (event.type != ButtonPress);
    XUngrabPointer(dpy, CurrentTime);

    focused_window = event.xbutton.subwindow;
    if (focused_window == None)
      focused_window = root;
    else
      focused_window = XmuClientWindow(dpy, focused_window);
    if (appres.debug)
      fprintf(stderr, "Selected window is: 0x%lX\n", focused_window);

    focused_subwindow = focused_window;
    do {  /* search the child window */
      XQueryPointer(dpy, focused_subwindow, &root, &child,
                    &junk_i, &junk_i, &junk_i, &junk_i, &junk_u);
      if (child != None) {
        focused_subwindow = child;
        if (appres.debug)
          fprintf(stderr, "  going down: 0x%lX\n", focused_subwindow);
      }
    } while (child != None);
    if (appres.kterm) {
      if (focused_children != NULL) XFree(focused_children);
      XQueryTree(dpy, focused_subwindow, &root, &parent,
                 &focused_children, &n_focused_children);
      if (appres.debug)
        fprintf(stderr, "Window 0x%lX has %d children\n",
                focused_subwindow, n_focused_children);
    }
    if (appres.list_widgets || strlen(appres.widget) != 0) {
      child = FindWidget(toplevel, focused_window, appres.widget);
      if (child != None) focused_subwindow = child;
    }
  } else {
    fprintf(stderr, "%s: cannot grab pointer", PROGRAM_NAME);
  }
}

/*
 * Read keyboard mapping and modifier mapping.
 * Keyboard mapping is used to know what keys are in shifted position.
 * Modifier mapping is required because we should know Alt and Meta
 * key are used as which modifier.
 */
static void ReadKeymap(void)
{
  int i, j;
  int keycode;
  KeySym keysym;

  XDisplayKeycodes(dpy, &min_keycode, &max_keycode);
  keysym_table = XGetKeyboardMapping(dpy,
                             min_keycode, max_keycode - min_keycode + 1,
                             &keysym_per_keycode);

  modifiers = XGetModifierMapping(dpy);
  for (i = 0; i < 8; i++) {
    for (j = 0; j < modifiers->max_keypermod; j++) {
      keycode = modifiers->modifiermap[i * modifiers->max_keypermod + j];
      keysym = keysym_table[(keycode - min_keycode) * keysym_per_keycode];
      if (keysym == XK_Alt_L || keysym == XK_Alt_R) {
        alt_mask = 1 << i;
      } else if (keysym == XK_Meta_L || keysym == XK_Meta_R) {
        meta_mask = 1 << i;
      } else if (keysym == XK_Mode_switch) {
        mode_sw_mask = 0x0101 << i;
        /* I don't know why, but 0x2000 was required for mod3 on my Linux box */
      }
    }
  }
}

/*
 * This will called when X error is detected when attempting to
 * send a event to a client window;  this will notmally caused
 * when the client window is destroyed.
 */
static int MyErrorHandler(Display *my_dpy, XErrorEvent *event)
{
  char msg[200];

  error_detected = TRUE;
  if (event->error_code == BadWindow) {
    /* ignore the error if it was BadWindow for XSetInputFocus() XSendEvent() */
    if (appres.debug) {
      switch (event->request_code) {
      case X_SendEvent:
        if (appres.debug)
          fprintf(stderr, "%s: couldn't send event to window 0x%lX\n",
                  PROGRAM_NAME, (long)focused_window);
        return 1;
      case X_SetInputFocus:
        if (appres.debug)
          fprintf(stderr, "%s: couldn't set input focus to window 0x%lX\n",
                  PROGRAM_NAME, (long)focused_window);
        return 1;
      }
    }
  }
  XGetErrorText(my_dpy, event->error_code, msg, sizeof(msg) - 1);
  fprintf(stderr, "X error trapped: %s\n", msg);
  return 0;
}

/*
 * Send event to the focused window.
 * If input focus is specified explicitly, select the window
 * before send event to the window.
 */
static void SendEvent(XKeyEvent *event)
{
  XSync(event->display, FALSE);
  XSetErrorHandler(MyErrorHandler);

  error_detected = FALSE;
  if (focused_window != None) {
    /* set input focus if input focus is set explicitly */
    if (appres.debug)
      fprintf(stderr, "Set input focus to window 0x%lX (0x%lX)\n",
              (long)focused_window, (long)event->window);
    XSetInputFocus(event->display, focused_window, RevertToParent, CurrentTime);
    XSync(event->display, FALSE);
  }
  if (!error_detected) {
    XSendEvent(event->display, event->window, TRUE, KeyPressMask, (XEvent *)event);
    XSync(event->display, FALSE);

    if (error_detected
        && (focused_subwindow != None) && (focused_subwindow != event->window)) {
      error_detected = FALSE;
      event->window = focused_subwindow;
      if (appres.debug)
        fprintf(stderr, "   retry: send event to window 0x%lX (0x%lX)\n",
                (long)focused_window, (long)event->window);
      XSendEvent(event->display, event->window, TRUE, KeyPressMask, (XEvent *)event);
      XSync(event->display, FALSE);
    }
  }

  if (error_detected) {
    /* reset focus because focused window is (probably) no longer exist */
    focused_window = None;
    focused_subwindow = None;
  }

  XSetErrorHandler(NULL);
}

/*
 * Send sequence of KeyPressed/KeyReleased events to the focused
 * window to simulate keyboard.  If modifiers (shift, control, etc)
 * are set ON, many events will be sent.
 */
static void SendKeyPressedEvent(KeySym keysym, unsigned int shift)
{
  static Window parent = None;
  Window cur_focus;
  int revert_to;
  XKeyEvent event;
  int keycode;
  Window root, *children;
  unsigned int n_children;
  int inx;
  Boolean found;

  if (focused_subwindow != None)
    cur_focus = focused_subwindow;
  else
    XGetInputFocus(dpy, &cur_focus, &revert_to);

  if (appres.debug) {
    char ch = '?';
    if ((keysym & ~0x7f) == 0 && isprint(keysym)) ch = keysym;
    fprintf(stderr, "SendKeyPressedEvent: focus=0x%lX, key=0x%lX (%c), shift=0x%lX\n",
            (long)cur_focus, (long)keysym, ch, (long)shift);
  }

  if (appres.kterm && focused_subwindow != None) {
    XQueryTree(dpy, cur_focus, &root, &parent, &children, &n_children);
    if (n_focused_children < n_children) {
      /* kterm hack - if new child window exist,
         we assume it is window for IM and redirect events to the window */
      cur_focus = children[n_children - 1];
      if (appres.debug)
        fprintf(stderr, "SendKeyPressedEvent: (kterm) focus new child winow 0x%lX\n",
                (long)cur_focus);
    }
    XFree(children);
  }

  if (XtWindow(toplevel) != None) {
    if (toplevel_parent == None) {
      XQueryTree(dpy, XtWindow(toplevel),
                 &root, &toplevel_parent, &children, &n_children);
      XFree(children);
    }
    if (cur_focus == XtWindow(toplevel) || cur_focus == toplevel_parent) {
      /* notice user when the xvkbd window is focused */
      XBell(dpy, 0);
      return;
    }
  }

  found = FALSE;
  keycode = 0;
  if (keysym != NoSymbol) {
    for (keycode = min_keycode; !found && (keycode <= max_keycode); keycode++) {
      /* Determine keycode for the keysym:  we use this instead
         of XKeysymToKeycode() because we must know shift_state, too */
      inx = (keycode - min_keycode) * keysym_per_keycode;
      if (keysym_table[inx] == keysym) {
        shift &= ~mode_sw_mask;
        if (keysym_table[inx + 1] != 0) shift &= ~ShiftMask;
        found = TRUE;
        break;
      } else if (keysym_table[inx + 1] == keysym) {
        shift &= ~mode_sw_mask;
        shift |= ShiftMask;
        found = TRUE;
        break;
      }
    }
    if (!found && mode_sw_mask && 3 <= keysym_per_keycode) {
      for (keycode = min_keycode; !found && (keycode <= max_keycode); keycode++) {
        inx = (keycode - min_keycode) * keysym_per_keycode;
        if (keysym_table[inx + 2] == keysym) {
          shift &= ~ShiftMask;
          shift |= mode_sw_mask;
          found = TRUE;
          break;
        } else if (4 <= keysym_per_keycode && keysym_table[inx + 3] == keysym) {
          shift |= ShiftMask | mode_sw_mask;
          found = TRUE;
          break;
        }
      }
    }
    if (appres.debug) {
      fprintf(stderr, "SendKeyPressedEvent: keycode=0x%lX, shift=0x%lX\n",
              (long)keycode, (long)shift);
    }
  }

  event.display = dpy;
  event.window = cur_focus;
  event.root = XtScreen(toplevel)->root;
  event.subwindow = None;
  event.time = CurrentTime;
  event.x = 1;
  event.y = 1;
  event.x_root = 1;
  event.y_root = 1;
  event.same_screen = TRUE;

  event.type = KeyPress;
  event.state = 0;
  if (shift & ShiftMask) {
    event.keycode = XKeysymToKeycode(dpy, XK_Shift_L);
    SendEvent(&event);
    event.state |= ShiftMask;
  }
  if (shift & ControlMask) {
    event.keycode = XKeysymToKeycode(dpy, XK_Control_L);
    SendEvent(&event);
    event.state |= ControlMask;
  }
  if (shift & alt_mask) {
    event.keycode = XKeysymToKeycode(dpy, XK_Alt_L);
    SendEvent(&event);
    event.state |= alt_mask;
  }
  if (shift & meta_mask) {
    event.keycode = XKeysymToKeycode(dpy, XK_Meta_L);
    SendEvent(&event);
    event.state |= meta_mask;
  }
  if (shift & mode_sw_mask) {
    event.keycode = XKeysymToKeycode(dpy, XK_Mode_switch);
    SendEvent(&event);
    event.state |= mode_sw_mask;
  }

  if (keysym != NoSymbol) {  /* send event for the key itself */
    event.keycode = found ? keycode : XKeysymToKeycode(dpy, keysym);
    if (event.keycode == NoSymbol) {
      if (keysym < 0x80 && isprint(keysym))
        fprintf(stderr, "%s: no such key: %c\n",
                PROGRAM_NAME, (char)keysym);
      else if (XKeysymToString(keysym) != NULL)
        fprintf(stderr, "%s: no such key: keysym=%s (0x%lX)\n",
                PROGRAM_NAME, XKeysymToString(keysym), (long)keysym);
      else
        fprintf(stderr, "%s: no such key: keysym=0x%lX\n",
                PROGRAM_NAME, (long)keysym);
      XBell(dpy, 0);
    } else {
      SendEvent(&event);
      event.type = KeyRelease;
      SendEvent(&event);
    }
  }

  if (shift & ShiftMask) {
    event.keycode = XKeysymToKeycode(dpy, XK_Shift_L);
    SendEvent(&event);
    event.state &= ~ShiftMask;
  }
  if (shift & ControlMask) {
    event.keycode = XKeysymToKeycode(dpy, XK_Control_L);
    SendEvent(&event);
    event.state &= ~ControlMask;
  }
  if (shift & alt_mask) {
    event.keycode = XKeysymToKeycode(dpy, XK_Alt_L);
    SendEvent(&event);
    event.state &= ~alt_mask;
  }
  if (shift & meta_mask) {
    event.keycode = XKeysymToKeycode(dpy, XK_Meta_L);
    SendEvent(&event);
    event.state &= ~meta_mask;
  }
  if (shift & mode_sw_mask) {
    event.keycode = XKeysymToKeycode(dpy, XK_Mode_switch);
    SendEvent(&event);
    event.state &= ~mode_sw_mask;
  }
}

/*
 * Send given string to the focused window as if the string
 * is typed from a keyboard.
 */
static void KeyPressed(Widget w, char *key, char *data);

static void SendString(char *str)
{
  char *cp, *cp2;
  char key[30];
  int len;

  for (cp = str; *cp != '\0'; cp++) {
    if (*cp == '\\') {
      cp++;
      switch (*cp) {
      case '\0':
        fprintf(stderr, "%s: missing character after \"\\\"\n",
                PROGRAM_NAME);
        return;
      case '[':  /* we can write any keysym as "\[the-keysym]" here */
        cp2 = strchr(cp, ']');
        if (cp2 == NULL) {
          fprintf(stderr, "%s: no closing \"]\" after \"\\[\"\n",
                  PROGRAM_NAME);
        } else {
          len = cp2 - cp - 1;
          if (sizeof(key) <= len) len = sizeof(key) - 1;
          strncpy(key, cp + 1, len);
          key[len] = '\0';
          KeyPressed(None, key, NULL);
          cp = cp2;
        }
        break;
      case 'S': shift_state |= ShiftMask; break;
      case 'C': shift_state |= ControlMask; break;
      case 'A': shift_state |= alt_mask; break;
      case 'M': shift_state |= meta_mask; break;
      case 'b': SendKeyPressedEvent(XK_BackSpace, 0); break;
      case 't': SendKeyPressedEvent(XK_Tab, 0); break;
      case 'n': SendKeyPressedEvent(XK_Linefeed, 0); break;
      case 'r': SendKeyPressedEvent(XK_Return, 0); break;
      case 'e': SendKeyPressedEvent(XK_Escape, 0); break;
      case 'd': SendKeyPressedEvent(XK_Delete, 0); break;
      default: SendKeyPressedEvent(*cp, 0); break;
      }
    } else {
      SendKeyPressedEvent(*cp, shift_state);
      shift_state = 0;
    }
  }
}

/*
 * Highlight/unhighligh spcified modifier key on the screen.
 */
static void Highlight(char *name, Boolean state)
{
  char name1[20];
  Widget w;

  sprintf(name1, "*%s", name);
  w = XtNameToWidget(toplevel, name1);
  if (w != None) {
    if (strcmp(name, "Focus") == 0) {
      if (state) {
        XtVaSetValues(w, XtNforeground, appres.highlight_foreground, NULL);
      } else {
        XtVaSetValues(w, XtNforeground, appres.special_foreground, NULL);
      }
    } else {
      if (state) {
        XtVaSetValues(w, XtNbackground, appres.highlight_background,
                      XtNforeground, appres.highlight_foreground, NULL);
      } else {
        XtVaSetValues(w, XtNbackground, appres.special_background,
                      XtNforeground, appres.special_foreground, NULL);
      }
    }
  }
}

/*
 * Highlight/unhighligh keys on the screen to reflect the state.
 */
static void RefreshShiftState(void)
{
  static int last_shift = 0;
  static long last_focus = 0;
  int cur_shift;
  int changed;

  cur_shift = shift_state | mouse_shift;
  changed = cur_shift ^ last_shift;
  last_shift = cur_shift;
  if (changed & ShiftMask) {
    Highlight("Shift_L", cur_shift & ShiftMask);
    Highlight("Shift_R", cur_shift & ShiftMask);
  }
  if (changed & ControlMask) {
    Highlight("Control_L", cur_shift & ControlMask);
    Highlight("Control_R", cur_shift & ControlMask);
  }
  if (changed & alt_mask) {
    Highlight("Alt_L", cur_shift & alt_mask);
    Highlight("Alt_R", cur_shift & alt_mask);
  }
  if (changed & meta_mask) {
    Highlight("Meta_L", cur_shift & meta_mask);
    Highlight("Meta_R", cur_shift & meta_mask);
  }
  if (changed & LockMask) {
    Highlight("Caps_Lock", cur_shift & LockMask);
  }
  if (changed & mode_sw_mask) {
    Highlight("Mode_switch", cur_shift & mode_sw_mask);
  }
  if (last_focus != focused_window) {
    Highlight("Focus", focused_window != 0);
    last_focus = focused_window;
  }
}

/*
 * This function will be called when mouse button is pressed on a key
 * on the screen.  Most operation will be performed in KeyPressed()
 * which will be called as callback for the Command widgets, and we
 * only need to check which mouse button is pressed here.
 */
static void ButtonDownAction(Widget w, XEvent *event, String *pars, Cardinal *n_pars)
{
  switch (event->xbutton.button) {
  case Button2:
    mouse_shift |= ControlMask;
    break;
  case Button3:
  case Button4:
    mouse_shift |= ShiftMask;
    break;
  }
  RefreshShiftState();
}

/*
 * This function will be called when mouse button is released on a key
 * on the screen, after callback is called.
 */
static void ButtonUpAction(Widget w, XEvent *event, String *pars, Cardinal *n_pars)
{
  mouse_shift = 0;
  RefreshShiftState();
}

/*
 * This will be called when user pressed a key on the screen.
 */
static void KeyPressed(Widget w, char *key, char *data)
{
  int row, col;
  int cur_shift;
  char *key1;
  KeySym keysym;
  Boolean shifted;

  if (strncmp(key, "Shift", strlen("Shift")) == 0) {
    if (shift_state & ShiftMask) SendKeyPressedEvent(NoSymbol, shift_state);
    shift_state ^= ShiftMask;
  } else if (strncmp(key, "Control", strlen("Control")) == 0) {
    if (shift_state & ControlMask) SendKeyPressedEvent(NoSymbol, shift_state);
    shift_state ^= ControlMask;
  } else if (strncmp(key, "Alt", strlen("Alt")) == 0) {
    if (shift_state & alt_mask) SendKeyPressedEvent(NoSymbol, shift_state);
    shift_state ^= alt_mask;
  } else if (strncmp(key, "Meta", strlen("Meta")) == 0) {
    if (shift_state & meta_mask) SendKeyPressedEvent(NoSymbol, shift_state);
    shift_state ^= meta_mask;
  } else if (strcmp(key, "Caps_Lock") == 0) {
    if (shift_state & LockMask) SendKeyPressedEvent(NoSymbol, shift_state);
    shift_state ^= LockMask;
  } else if (strcmp(key, "Mode_switch") == 0) {
    if (shift_state & mode_sw_mask) SendKeyPressedEvent(NoSymbol, shift_state);
    shift_state ^= mode_sw_mask;
  } else {
    cur_shift = shift_state | mouse_shift;
    shifted = (cur_shift & ShiftMask);
    if (sscanf(key, "pad%d,%d", &row, &col) == 2) {
      key1 = shifted ? keypad_shift[row][col]: keypad[row][col];
    } else if (sscanf(key, "%d,%d", &row, &col) == 2) {
      if ((cur_shift & LockMask)
          && isalpha(keys_normal[row][col][0]) && keys_normal[row][col][1] == '\0')
        shifted = !shifted;
      if ((cur_shift & mode_sw_mask) && keys_altgr[row][col] != NULL) {
        if (shifted && keys_shift_altgr[row][col] != NULL)
          key1 = keys_shift_altgr[row][col];
        else
          key1 = keys_altgr[row][col];
      } else {
        if (shifted && keys_shift[row][col] != NULL)
          key1 = keys_shift[row][col];
        else
          key1 = keys_normal[row][col];
      }
    } else {
      key1 = key;
    } 
    if (strcmp(key1, "Focus") == 0) {
      if (cur_shift & ShiftMask) {
        focused_window = None;
        focused_subwindow = None;
      } else {
        GetFocusedWindow();
      }
    } else {
      if (strlen(key1) == 1) {
        SendKeyPressedEvent((KeySym)*key1 & 0xff, cur_shift);
      } else {
        keysym = XStringToKeysym(key1);
        if ((!appres.keypad_keysym && strncmp(key1, "KP_", 3) == 0)
            || XKeysymToKeycode(dpy, keysym) == NoSymbol) {
          switch ((unsigned)keysym) {
          case XK_KP_Equal: keysym = XK_equal; break;
          case XK_KP_Divide: keysym = XK_slash; break;
          case XK_KP_Multiply: keysym = XK_asterisk; break;
          case XK_KP_Add: keysym = XK_plus; break;
          case XK_KP_Subtract: keysym = XK_minus; break;
          case XK_KP_Enter: keysym = XK_Return; break;
          case XK_KP_1: keysym = XK_1; break;
          case XK_KP_2: keysym = XK_2; break;
          case XK_KP_3: keysym = XK_3; break;
          case XK_KP_4: keysym = XK_4; break;
          case XK_KP_5: keysym = XK_5; break;
          case XK_KP_6: keysym = XK_6; break;
          case XK_KP_7: keysym = XK_7; break;
          case XK_KP_8: keysym = XK_8; break;
          case XK_KP_9: keysym = XK_9; break;
          case XK_Shift_L: keysym = XK_Shift_R; break;
          case XK_Shift_R: keysym = XK_Shift_L; break;
          case XK_Control_L: keysym = XK_Control_R; break;
          case XK_Control_R: keysym = XK_Control_L; break;
          case XK_Alt_L: keysym = XK_Alt_R; break;
          case XK_Alt_R: keysym = XK_Alt_L; break;
          case XK_Meta_L: keysym = XK_Meta_R; break;
          case XK_Meta_R: keysym = XK_Meta_L; break;
          default: fprintf(stderr, "%s: no such key: %s\n",
                           PROGRAM_NAME, key1); break;
          }
        }
        SendKeyPressedEvent(keysym, cur_shift);
      }

      if ((cur_shift & ControlMask) && (cur_shift & alt_mask)) {
        if (strstr(XServerVendor(dpy), "XFree86") != NULL) {
          if (strcmp(key1, "KP_Add") == 0) {
            system("xvidtune -next");
          } else if (strcmp(key1, "KP_Subtract") == 0) {
            system("xvidtune -prev");
          }
        }
      }
    }
    shift_state = shift_state & LockMask;
  }
  RefreshShiftState();
}

/*
 * Redefine keyboard layout.
 * "spec" is a sequence of words separated with spaces, and it is
 * usally specified in app-defaults file, as:
 * 
 *   xvkbd.AltGrKeys: \
 *      F1 F2 F3 F4 F5 F6 F7 F8 F9 F10 F11 F12 BackSpace \n\
 *      Escape \271 \262 \263 \243 \254 \251 { [ ] } \\ ' ^ ' \n\
 *      ...
 *
 * White spaces separate the keys, and " \n" (note that white space
 * before the \n) separate the rows of keys.
 */
static void RedefineKeys(char *array[NUM_KEY_ROWS][NUM_KEY_COLS], const char *spec)
{
  char *s = XtNewString(spec);
  char *cp;
  int row, col;

  for (row = 0; row < NUM_KEY_ROWS; row++) {
    for (col = 0; col < NUM_KEY_COLS; col++) array[row][col] = NULL;
  }
  row = 0;
  col = 0;
  cp = strtok(s, " ");
  while (cp != NULL) {
    if (*cp == '\n') {
      row = row + 1;
      col = 0;
      cp = cp + 1;
    }
    if (*cp != '\0') {
      if (NUM_KEY_ROWS <= row) {
        fprintf(stderr, "%s: too many key rows: \"%s\" ignored\n",
                PROGRAM_NAME, cp);
      } else if (NUM_KEY_COLS <= col) {
        fprintf(stderr, "%s: too many keys in a row: \"%s\" ignored\n",
                PROGRAM_NAME, cp);
      } else {
        array[row][col] = XtNewString(cp);
        col = col + 1;
      }
    }
    cp = strtok(NULL, " ");
  }
  XtFree(s);
}

/*
 * Create keyboard on the screen.
 */
static Widget MakeKey(Widget parent, const char *name, const char *label, Pixel color)
{
  static Pixmap up_pixmap = None;
  static Pixmap down_pixmap = None;
  static Pixmap left_pixmap = None;
  static Pixmap right_pixmap = None;
  Widget w;
  Window scr = XtScreen(parent)->root;
  char str[50];
  int len;

  if (!appres.auto_repeat
      || strncmp(name, "Shift", strlen("Shift")) == 0
      || strncmp(name, "Control", strlen("Control")) == 0
      || strncmp(name, "Alt", strlen("Alt")) == 0
      || strncmp(name, "Meta", strlen("Meta")) == 0
      || strcmp(name, "Caps_Lock") == 0
      || strcmp(name, "Mode_switch") == 0
      || strcmp(name, "Focus") == 0) {
    w = XtVaCreateManagedWidget(name, commandWidgetClass, parent,
                                XtNbackground, color, NULL);
  } else {
    w = XtVaCreateManagedWidget(name, repeaterWidgetClass, parent,
                                XtNbackground, color, NULL);
  }
  XtAddCallback(w, XtNcallback, (XtCallbackProc)KeyPressed, (XtPointer)name);

  if (label != NULL) {
    strncpy(str, label, sizeof(str) - 1);
    if (strcmp(str, "space") == 0) strcpy(str, "");
    len = strlen(str);
    if (3 <= len) {
      if (str[1] == '_') str[1] = ' ';
      if (str[len - 2] == '_') str[len - 2] = ' ';
    }
    XtVaSetValues(w, XtNlabel, str, NULL);

    if (strcmp(label, "up") == 0) {
      if (up_pixmap == None)
        up_pixmap = XCreateBitmapFromData(dpy, scr,
                                (char *)up_bits, up_width, up_height);
      XtVaSetValues(w, XtNbitmap, up_pixmap, NULL);
    } else if (strcmp(label, "down") == 0) {
      if (down_pixmap == None)
        down_pixmap = XCreateBitmapFromData(dpy, scr,
                                (char *)down_bits, down_width, down_height);
      XtVaSetValues(w, XtNbitmap, down_pixmap, NULL);
    } else if (strcmp(label, "left") == 0) {
      if (left_pixmap == None)
        left_pixmap = XCreateBitmapFromData(dpy, scr,
                                (char *)left_bits, left_width, left_height);
      XtVaSetValues(w, XtNbitmap, left_pixmap, NULL);
    } else if (strcmp(label, "right") == 0) {
      if (right_pixmap == None)
        right_pixmap = XCreateBitmapFromData(dpy, scr,
                                (char *)right_bits, right_width, right_height);
      XtVaSetValues(w, XtNbitmap, right_pixmap, NULL);
    }
  }

  return w;
}

static void MakeKeyboard(void)
{
  Widget form, key, left, upper;
  Pixel color;
  XFontStruct *font;
  Dimension wd, max_wd;
  int row, col, first_row;
  char name[20];
  Widget key_box[NUM_KEY_ROWS];
  Widget keypad_box;
  Widget keypad_row[NUM_KEYPAD_ROWS];

  form = XtVaCreateManagedWidget("form", formWidgetClass, toplevel, NULL);

  key_box[0] = None;
  key_box[1] = None;
  first_row = appres.function_key ? 0 : 1;
  if (!appres.keypad_only) {
    for (row = first_row; row < NUM_KEY_ROWS; row++) {
      sprintf(name, "row%d", row);
      key_box[row] = XtVaCreateManagedWidget(name, formWidgetClass, form, NULL);
      if (row != first_row)
        XtVaSetValues(key_box[row], XtNfromVert, key_box[row - 1], NULL);
      else if (!appres.function_key)
        XtVaSetValues(key_box[row], XtNvertDistance, 0, NULL);
        
      left = None;
      for (col = 0; keys_normal[row][col] != NULL; col++) {
        strcpy(name, keys_normal[row][col]);
        if ('A' <= name[0] && name[0] <= 'Z') {
          /* on SunOS 5.6, isupper() reported odiaeresis is an upper case letter... */
          color = appres.special_background;
          if (key_labels[row][col] != NULL
              && strchr(key_labels[row][col], '\n') != NULL) font = appres.keypad_font;
          else font = appres.special_font;
        } else {
          color = appres.general_background;
          font = appres.general_font;
          if (isalpha(name[0])) font = appres.letter_font;
          if (strcmp(name, "space") != 0) sprintf(name, "%d,%d", row, col);
        }
        key = MakeKey(key_box[row], XtNewString(name), key_labels[row][col], color);
        XtVaGetValues(key, XtNwidth, &wd, NULL);
        if (wd <= 1) {
          /* keys can be removed by setting its width to 1 */
          XtDestroyWidget(key);
        } else {
          XtVaSetValues(key, XtNfont, font, NULL);
          if (left != None) XtVaSetValues(key, XtNfromHoriz, left, NULL);
          left = key;
          if (strcmp(keys_normal[row][col], "Multi_key") == 0
              && XKeysymToKeycode(dpy, XK_Multi_key) == NoSymbol) {
            XtVaSetValues(key, XtNsensitive, FALSE, NULL);
          } else if (strcmp(keys_normal[row][col], "Mode_switch") == 0
              && mode_sw_mask == 0) {
            XtVaSetValues(key, XtNsensitive, FALSE, NULL);
          }
        }
      }
    }
  }

  if (appres.keypad) {
    keypad_box = XtVaCreateManagedWidget("keypad", formWidgetClass, form, NULL);
    if (!appres.keypad_only)
      XtVaSetValues(keypad_box, XtNfromHoriz, key_box[1], NULL);
    else
      XtVaSetValues(keypad_box, XtNhorizDistance, 0, NULL);
    if (key_box[0] != None)
      XtVaSetValues(keypad_box, XtNfromVert, key_box[0], NULL);
    else
      XtVaSetValues(keypad_box, XtNvertDistance, 0, NULL);
    upper = None;
    for (row = 0; row < NUM_KEYPAD_ROWS; row++) {
      left = None;
      for (col = 0; keypad[row][col] != NULL; col++) {
        font = appres.keypad_font;
        if (strlen(keypad_label[row][col]) == 1) font = appres.letter_font;
        color = appres.special_background;
        if (strcmp(keypad[row][col], "Focus") == 0)
          color = appres.focus_background;
        strcpy(name, keypad[row][col]);
        if (strcmp(name, "Focus") != 0)
          sprintf(name, "pad%d,%d", row, col);
        key = MakeKey(keypad_box, XtNewString(name),
                      keypad_label[row][col], color);
        XtVaSetValues(key, XtNfont, font, NULL);
        if (row != 0)
          XtVaSetValues(key, XtNfromVert, keypad_row[row - 1], NULL);
        if (left != None)
          XtVaSetValues(key, XtNfromHoriz, left, NULL);
        if (col == 0)
          keypad_row[row] = key;
        left = key;
      }
    }
  }

  if (!appres.keypad_only && appres.function_key && appres.keypad) {
    XtVaCreateManagedWidget("banner", labelWidgetClass, form,
                            XtNfromHoriz, key_box[1],
                            XtNlabel, PROGRAM_NAME_WITH_VERSION, NULL);
  }

  XtRealizeWidget(toplevel);
  if (!appres.debug && key_box[first_row] != None) {
    if (appres.keypad) {
      XtVaGetValues(key_box[1], XtNwidth, &max_wd, NULL);
    } else {
      max_wd = 0;
      for (row = first_row; row < NUM_KEY_ROWS; row++) {
        XtVaGetValues(key_box[row], XtNwidth, &wd, NULL);
        if (max_wd < wd) max_wd = wd;
      }
    }
    for (row = first_row; row < NUM_KEY_ROWS; row++) {
      XtVaSetValues(key_box[row], XtNwidth, max_wd, NULL);
    }
  }
  if (0 < strlen(appres.geometry)) {
    XtVaSetValues(toplevel, XtNgeometry, appres.geometry, NULL);
    XtUnrealizeWidget(toplevel);
    XtRealizeWidget(toplevel);
  }
  XtMapWidget(toplevel);
}

/*
 * WM_DELETE_WINDOW has been sent - terminate the program.
 */
static void DeleteWindowProc(Widget w, XEvent *event,
                             String *pars, Cardinal *n_pars)
{
  XtDestroyApplicationContext(XtWidgetToApplicationContext(toplevel));
  exit(0);
}

/*
 * The main program.
 */
void main(int argc, char *argv[])
{
  static XtActionsRec actions[] = {
    { "DeleteWindowProc", DeleteWindowProc },
    { "ButtonDownAction", ButtonDownAction },
    { "ButtonUpAction", ButtonUpAction },
  };
  static String fallback_resources[] = {
    "?.title: xvkbd - Virtual Keyboard",
    "?.iconName: xvkbd",
    "?.input: false",
    "?.mappedWhenManaged: false",
    "?.translations: #override \
        <Message>WM_PROTOCOLS: DeleteWindowProc()",
    "?.Form.translations: <BtnDown>:",
    "?.Form.?.translations: <BtnDown>:",
    "?*Command.translations: #override \\n\
        <BtnDown>: ButtonDownAction() set()\\n\
        <BtnUp>: notify() unset() ButtonUpAction()",
    "?*Repeater.translations: #override \\n\
        <BtnDown>: ButtonDownAction() set() start() \\n\
        <BtnUp>: stop() unset() ButtonUpAction()",
    "?*Repeater.initialDelay: 600",
    "?*Repeater.minimumDelay: 100",
    "?*Repeater.decay: 500",
    "?*generalFont: -*-lucidatypewriter-bold-r-*-*-12-*-*-*-*-*-iso8859-*",
    "?*letterFont: -*-helvetica-bold-r-*-*-12-*-*-*-*-*-iso8859-*",
    "?*specialFont: -*-helvetica-medium-r-*-*-12-*-*-*-*-*-iso8859-*",
    "?*keypadFont: -*-helvetica-medium-r-*-*-10-*-*-*-*-*-iso8859-*",
    "?*banner.font: -*-lucida-bold-i-*-*-14-*-*-*-*-*-iso8859-*",
    "?*beNiceToColormap: false",
    "?*topShadowContrast: 40",
    "?*bottomShadowContrast: 80",
    "?*shadowWidth: 2",
    "?*Label.shadowWidth: 0",
    "?*Background: gray85",
    "?*specialBackground: gray75",
    "?*focusBackground: bisque2",
    "?*Foreground: black",
    "?*highlightBackground: gray90",
    "?*highlightForeground: forestgreen",
    "?*highlightThickness: 1",
    "?*Form.resizable: true",
    "?*defaultDistance: 0",
    "?*Form.borderWidth: 0",
    "?*Label.borderWidth: 0",
    "?*Command.internalWidth: 1",
    "?*Repeater.internalWidth: 1",
    "?*Command.resize: false",
    "?*Repeater.resize: false",
    "?*Command.justify: center",
    "?*Repeater.justify: center",

    "?*banner.width: 150",
    "?*banner.justify: right",

    "?*row1.vertDistance: 5",
    "?*F5.horizDistance: 5",
    "?*F9.horizDistance: 5",
    "?*BackSpace.horizDistance: 5",

    "?*Command.height: 30",
    "?*Repeater.height: 30",
    "?*row0.Repeater.height: 25",
    "?*Command.width: 30",
    "?*Repeater.width: 30",
    "?*space.width: 115",
    "?*Tab.width: 45",
    "?*Control_L.width: 60",
    "?*Shift_L.width: 75",
    "?*Shift_R.width: 40",
    "?*Multi_key.width: 35",
    "?*Caps_Lock.width: 35",
    "?*Alt_L.width: 45",
    "?*Alt_R.width: 45",
    "?*Meta_L.width: 45",
    "?*Meta_R.width: 45",
    "?*BackSpace.width: 75",
    "?*Delete.width: 45",
    "?*Return.width: 60",

    "?*keypad.vertDistance: 5",
    "?*keypad.horizDistance: 10",
    "?*keypad*Command.width: 35",
    "?*keypad*Repeater.width: 35",
    "?*keypad*Command.justify: left",
    "?*keypad*Repeater.justify: left",
#ifdef XAW3D
    "?*keypad*pad3,3.height: 60",
    "?*keypad*pad4,0.width: 70",
#else
    "?*keypad*pad3,3.height: 62",
    "?*keypad*pad4,0.width: 72",
#endif
    NULL,
  };

  char ch;
  Window child;
  char *class_name;

  toplevel = XtVaAppInitialize(NULL, "XVkbd",
                               options, XtNumber(options),
                               &argc, argv, fallback_resources, NULL);
  dpy = XtDisplay(toplevel);
  app_con = XtWidgetToApplicationContext(toplevel);
  XtAppAddActions(app_con, actions, XtNumber(actions));

  if (1 < argc) {
    fprintf(stderr, "%s: illegal option: %s\n\n", PROGRAM_NAME, argv[1]);
  }

  XtGetApplicationResources(toplevel, &appres,
            application_resources, XtNumber(application_resources),
            NULL, 0);
  if (appres.compact) {
    appres.keypad = FALSE;
    appres.function_key = FALSE;
  }
  if (appres.keypad_only && !appres.keypad) {
    appres.keypad_only = FALSE;
    fprintf(stderr, "%s: warning: keypad_only ignored\n", PROGRAM_NAME);
  }
  if (0 < strlen(appres.window)) {
    if (sscanf(appres.window, "0x%lX%c", &focused_window, &ch) != 1) {
      if (sscanf(appres.window, "%ld%c", &focused_window, &ch) != 1) {
        focused_window = FindWindow(RootWindow(dpy, DefaultScreen(dpy)),
                                    appres.window);
        if (focused_window == None) {
          fprintf(stderr, "xvkbd: no such window: %s\n", appres.window);
        }
      }
    }
  }
  focused_subwindow = focused_window;

  ReadKeymap();
  if (strlen(appres.text) != 0) {
    if (focused_window != None &&
        (appres.list_widgets || strlen(appres.widget) != 0)) {
      XtVaSetValues(toplevel, XtNwidth, 1, XtNheight, 1, NULL);
      XtRealizeWidget(toplevel);
      child = FindWidget(toplevel, focused_window, appres.widget);
      if (child != None) focused_subwindow = child;
    }
    SendString(appres.text);
    exit(0);
  } else {
    if (0 < strlen(appres.keys_normal)) RedefineKeys(keys_normal, appres.keys_normal);
    if (0 < strlen(appres.keys_shift)) RedefineKeys(keys_shift, appres.keys_shift);
    if (0 < strlen(appres.keys_altgr)) RedefineKeys(keys_altgr, appres.keys_altgr);
    if (0 < strlen(appres.keys_shift_altgr)) RedefineKeys(keys_shift_altgr, appres.keys_shift_altgr);
    if (0 < strlen(appres.key_labels)) RedefineKeys(key_labels, appres.key_labels);
    MakeKeyboard();
    RefreshShiftState();
    wm_delete_window = XInternAtom(dpy, "WM_DELETE_WINDOW", FALSE);
    XSetWMProtocols(dpy, XtWindow(toplevel), &wm_delete_window, 1);
    if (focused_window != None &&
        (appres.list_widgets || strlen(appres.widget) != 0)) {
      child = FindWidget(toplevel, focused_window, appres.widget);
      if (child != None) focused_subwindow = child;
    }
    XtAppMainLoop(app_con);
  }
}
